"""Higher-level api for searching for papers, uses an object interface
with overridden magic methods for syntactic sugar
"""
from typing import Union

from sxolar.api import arxiv
from sxolar.api.arxiv import LogicalOperator, SearchField


class Query:
    """Represents a query clause for the arxiv API

    Attributes:
        value (str): The value to search for
        operator (str): The operator to use
    """

    def __init__(self, value: str):
        """Creates a new query

        Args:
            value:
                str, the value to search for
        """
        self.value = value

    def __str__(self):
        """Returns the string representation of the query"""
        return self.value

    def __and__(self, other):
        """Overloads the and operator to create a new query"""
        return self.and_(other)

    def __or__(self, other):
        """Overloads the or operator to create a new query"""
        return self.or_(other)

    def __sub__(self, other):
        """Overloads the subtraction operator to create a new query"""
        return self.and_not(other)

    def and_(self, other: Union[str, 'Query']):
        """Join two queries with the AND operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(f'{self}{LogicalOperator.AND}{other}')

    def and_not(self, other):
        """Join two queries with the AND NOT operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(f'{self}{LogicalOperator.AND_NOT}{other}')

    def or_(self, other):
        """Join two queries with the OR operator

        Args:
            other:
                str, the other query to join with

        Returns:
            Query: A new query object
        """
        return Query(f'{self}{LogicalOperator.OR}{other}')

    def wrap(self):
        """Wrap the query in parenthesis

        Returns:
            Query: A new query object
        """
        return Query(f'({self})')

    def search(self, start: int = 0, max_results: int = 10):
        """Searches the arxiv API with the query

        Args:
            start:
                int, optional, The starting index of the results
            max_results:
                int, optional, The maximum number of results to return

        Returns:
            list: A list of dictionaries representing the search results
        """
        return arxiv._query(self.value, id_list=None, start=start, max_results=max_results)


class Author(Query):
    """Represents an author query for the arxiv API
    """

    def __init__(self, name: str):
        """Creates a new author query

        Args:
            name:
                str, the name of the author, "First Last"
        """
        if not name.startswith(SearchField.AUTHOR):
            name = f'{SearchField.AUTHOR}:{name}'
        super().__init__(name)


class Title(Query):
    """Represents a title query for the arxiv API
    """

    def __init__(self, title: str):
        """Creates a new title query

        Args:
            title:
                str, the title of the paper
        """
        if not title.startswith(SearchField.TITLE):
            title = f'{SearchField.TITLE}:{title}'
        super().__init__(title)


class Abstract(Query):
    """Represents an abstract query for the arxiv API
    """

    def __init__(self, abstract: str):
        """Creates a new abstract query

        Args:
            abstract:
                str, the abstract of the paper
        """
        if not abstract.startswith(SearchField.ABSTRACT):
            abstract = f'{SearchField.ABSTRACT}:{abstract}'
        super().__init__(abstract)


class All(Query):
    """Represents an all query for the arxiv API
    """

    def __init__(self, all_: str):
        """Creates a new all query

        Args:
            all_:
                str, the value to search for
        """
        if not all_.startswith(SearchField.ALL):
            all_ = f'{SearchField.ALL}:{all_}'
        super().__init__(all_)


class JournalRef(Query):
    """Represents a journal reference query for the arxiv API
    """

    def __init__(self, journal_ref: str):
        """Creates a new journal reference query

        Args:
            journal_ref:
                str, the journal reference
        """
        if not journal_ref.startswith(SearchField.JOURNAL_REFERENCE):
            journal_ref = f'{SearchField.JOURNAL_REFERENCE}:{journal_ref}'
        super().__init__(journal_ref)


class Category(Query):
    """Represents a category query for the arxiv API
    """

    def __init__(self, category: str):
        """Creates a new category query

        Args:
            category:
        """
        if not category.startswith(SearchField.CATEGORY):
            category = f'{SearchField.CATEGORY}:{category}'
        super().__init__(category)



"""Arxiv API wrappers for sxolar, low-level functions for querying the Arxiv API.
For more user-friendly functions, see the `sxolar.api.query` module.

References:
    [1] API Basics: https://info.arxiv.org/help/api/basics.html
    [2] Rate Limits: https://info.arxiv.org/help/api/tou.html
    [3] Search Query Language: https://info.arxiv.org/help/api/user-manual.html#query_details
    [4] Entry output format: https://info.arxiv.org/help/api/user-manual.html#_entry_metadata
"""
import collections
from typing import List, Union
from urllib import parse
from defusedxml import ElementTree as SecureElementTree
from xml.etree import ElementTree

from requests_ratelimiter import LimiterSession

# Impose the public ratelimit by default [2]
SESSION = LimiterSession(per_minute=20)

# Define the base URL for the Arxiv API [1]
URL_BASE = 'http://export.arxiv.org/api/'
URL_QUERY = URL_BASE + 'query'

# Define XML tags for the Arxiv API [4]
TAG_PREFIX = '{http://www.w3.org/2005/Atom}'
TAG_ENTRY = 'entry'
TAG_TITLE = 'title'
TAG_ID = 'id'
TAG_PUBLISHED = 'published'
TAG_UPDATED = 'updated'
TAG_SUMMARY = 'summary'
TAG_AUTHOR = 'author'
TAG_NAME = 'name'
TAG_AFFILIATION = 'affiliation'
TAG_CATEGORY = 'category'
TAG_TERM = 'term'
TAG_SCHEME = 'scheme'

# Define the Entry output formats for the Arxiv API [4]
Entry = collections.namedtuple('Entry', 'title id published updated summary author category')
Author = collections.namedtuple('Author', 'name affiliation')
Category = collections.namedtuple('Category', 'term scheme')


class SearchField:
    """Enumeration of search fields for the Arxiv API [3]
    """
    TITLE = 'ti'
    AUTHOR = 'au'
    ABSTRACT = 'abs'
    COMMENT = 'co'
    JOURNAL_REFERENCE = 'jr'
    CATEGORY = 'cat'
    REPORT_NUMBER = 'rn'
    ID = 'id'
    ALL = 'all'


class LogicalOperator:
    """Enumeration of logical operators for the Arxiv API [3]
    """
    AND = ' AND '
    OR = ' OR '
    AND_NOT = ' ANDNOT '


def find(entry: ElementTree.Element, tag: str, find_all: bool = False) -> Union[str, List[str]]:
    """Find the tag in the entry and return the text.

    Args:
        entry (ElementTree.Element): The entry to search.
        tag (str): The tag to search for.

    Returns:
        str: The text of the tag.
    """
    if not tag.startswith(TAG_PREFIX):
        tag = TAG_PREFIX + tag

    if find_all:
        return entry.findall(tag)

    res = entry.find(tag)
    if res is not None:
        return res.text


def parse_entry(entry: ElementTree.Element) -> Entry:
    """Parse an entry from the Arxiv API response.

    Args:
        entry (ElementTree.Element): The entry to parse.

    Returns:
        Entry: The parsed entry.
    """
    # Parse the authors
    authors = [Author(name=find(author, TAG_NAME),
                      affiliation=find(author, TAG_AFFILIATION)) for author in find(entry, TAG_AUTHOR, find_all=True)]

    # Parse the categories
    categories = [Category(term=category.attrib[TAG_TERM],
                           scheme=category.attrib[TAG_SCHEME]) for category in find(entry, TAG_CATEGORY, find_all=True)]

    # Return the parsed entry
    return Entry(title=find(entry, TAG_TITLE),
                 id=find(entry, TAG_ID),
                 published=find(entry, TAG_PUBLISHED),
                 updated=find(entry, TAG_UPDATED),
                 summary=find(entry, TAG_SUMMARY),
                 author=authors,
                 category=categories)


def get_and_parse(url: str, params: dict) -> List[Entry]:
    """Get and parse the response from the Arxiv API, the payloads
    are encoded using the Atom 1 XML format.

    Args:
        url (str): The endpoint to query
        params (dict): The parameters to pass to the query

    Returns:
        dict: The parsed response
    """
    # Get the response
    response = SESSION.get(url, params=params)

    # Parse the response securely into ElementTree
    root = SecureElementTree.fromstring(response.text)

    # TODO finish parsing response into a list of named tuples if no errors, otherwise raise the error
    if len(root) == 1 and root[0].tag == 'error':
        raise ValueError(f'No results found. Error: {root[0].text}')

    entries = [parse_entry(entry) for entry in find(root, TAG_ENTRY, find_all=True)]

    # Return the parsed response
    return entries


def _extend_query(query: str, field: SearchField, value: Union[str, List[str]],
                  how: LogicalOperator = LogicalOperator.AND, how_list: LogicalOperator = LogicalOperator.OR) -> str:
    """Extend the query with the given field and value.

    Args:
        query:
            str, The current query string.
        field:
            SearchField, The field to search in.
        value:
            Union[str, List[str]], The value to search for.
        how:
            LogicalOperator, The logical operator to use when adding the field.
        how_list:
            LogicalOperator, The logical operator to use when adding a list of values.

    Returns:
        str: The extended query string.
    """
    # Check if query exists, if so then extend with cross-field logical operator
    if query:
        query += how

    # Check if value is scalar or list, then extend query
    if isinstance(value, str):
        query += f'{field}:{value}'
    elif isinstance(value, list):
        query += f'({field}:{how_list.join(value)})'

    return query


def format_search_query(title: Union[str, List[str]] = None, author: Union[str, List[str]] = None, abstract: Union[str, List[str]] = None, comment: Union[str, List[str]] = None,
                        journal_reference: Union[str, List[str]] = None, category: Union[str, List[str]] = None, report_number: Union[str, List[str]] = None, id_list: List[str] = None,
                        all_: Union[str, List[str]] = None, how: LogicalOperator = LogicalOperator.AND, how_list: LogicalOperator = LogicalOperator.OR) -> Union[str, None]:
    """Format the search query for the Arxiv API.

    Args:
        title:
            Union[str, List[str]], optional, The title to search for. Defaults to None.
        author:
            Union[str, List[str]], optional, The author to search for. Defaults to None.
        abstract:
            Union[str, List[str]], optional, The abstract to search for. Defaults to None.
        comment:
            Union[str, List[str]], optional, The comment to search for. Defaults to None.
        journal_reference:
            Union[str, List[str]], optional, The journal reference to search for. Defaults to None.
        category:
            Union[str, List[str]], optional, The category to search for. Defaults to None.
        report_number:
            Union[str, List[str]], optional, The report number to search for. Defaults to None.
        id_list:
            List[str], optional, The list of Arxiv IDs to search for. Defaults to None.
        all_:
            Union[str, List[str]], optional, The all field to search for. Defaults to None.

    Returns:
        str or None: The formatted query string, or None if no fields are provided.
    """
    # Short-circuit if no fields are provided
    if all(v is None for v in (title, author, abstract, comment, journal_reference, category, report_number, id_list, all_)):
        return None

    query = ''

    for field, value in zip(
            (SearchField.TITLE, SearchField.AUTHOR, SearchField.ABSTRACT, SearchField.COMMENT, SearchField.JOURNAL_REFERENCE, SearchField.CATEGORY, SearchField.REPORT_NUMBER, SearchField.ID, SearchField.ALL),
            (title, author, abstract, comment, journal_reference, category, report_number, id_list, all_)
    ):
        if value is not None:
            query = _extend_query(query, field, value, how=how, how_list=how_list)

    return parse.quote(query, safe='/:&=')


def _query(search_query: str = None, id_list: List[str] = None, start: int = 0, max_results: int = 10) -> dict:
    """Query the Arxiv API with the given parameters.

    Args:
        search_query:
            str, optional, The query string to search for. Defaults to None.
        id_list:
            List[str], optional, A list of Arxiv IDs to search for. Defaults to None.
        start:
            int, optional, The index to start the search from. Defaults to 0.
        max_results:
            int, optional, The maximum number of results to return. Defaults to 10.

    Returns:
        dict: The parsed response from the API
    """
    # Define the parameters for the query
    params = {
        'search_query': search_query,
        'id_list': id_list,
        'start': start,
        'max_results': max_results
    }

    # Filter out the None values
    params = {k: v for k, v in params.items() if v is not None}

    # Get and parse the response
    return get_and_parse(URL_QUERY, params)


def query(title: Union[str, List[str]] = None, author: Union[str, List[str]] = None, abstract: Union[str, List[str]] = None, comment: Union[str, List[str]] = None,
          journal_reference: Union[str, List[str]] = None, category: Union[str, List[str]] = None, report_number: Union[str, List[str]] = None, id_list: List[str] = None,
          all_: Union[str, List[str]] = None, how: LogicalOperator = LogicalOperator.AND, how_list: LogicalOperator = LogicalOperator.OR, start: int = 0, max_results: int = 10) -> dict:
    """Query the Arxiv API with the given parameters.

    Args:
        title:
            Union[str, List[str]], optional, The title to search for. Defaults to None.
        author:
            Union[str, List[str]], optional, The author to search for. Defaults to None.
        abstract:
            Union[str, List[str]], optional, The abstract to search for. Defaults to None.
        comment:
            Union[str, List[str]], optional, The comment to search for. Defaults to None.
        journal_reference:
            Union[str, List[str]], optional, The journal reference to search for. Defaults to None.
        category:
            Union[str, List[str]], optional, The category to search for. Defaults to None.
        report_number:
            Union[str, List[str]], optional, The report number to search for. Defaults to None.
        id_list:
            List[str], optional, The list of Arxiv IDs to search for. Defaults to None.
        all_:
            Union[str, List[str]], optional, The all field to search for. Defaults to None.
        how:
            LogicalOperator, optional, The logical operator to use when adding the field. Defaults to LogicalOperator.AND.
        how_list:
            LogicalOperator, optional, The logical operator to use when adding a list of values. Defaults to LogicalOperator.OR.
        start:
            int, optional, The index to start the search from. Defaults to 0.
        max_results:
            int, optional, The maximum number of results to return. Defaults to 10.
    """
    # Format the search query
    search_query = format_search_query(title, author, abstract, comment, journal_reference, category, report_number, id_list, all_, how, how_list)

    # Short-circuit if no search query is provided
    if search_query is None and id_list is None:
        raise ValueError('No search query provided; cannot query the entire Arxiv.')

    # Query the API
    return _query(search_query, id_list, start, max_results)



"""Setup module for sxolar package
"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(name="sxolar",
      version="0.1.0",
      description="Scholar's tools for working with Arxiv",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/JWKennington/sxolar",
      author="J. W. Kennington",
      author_email="jameswkennington@gmail.com",
      classifiers=[
          "Development Status :: 3 - Alpha",
          # Pick your license as you wish
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Programming Language :: Python :: 3 :: Only",
      ],
      keywords="topic reference, bibliography, education, reference index",
      packages=find_packages(),
      python_requires=">=3.7, <4",
      install_requires=["pyyaml"],
      extras_require={  # Optional
          "dev": ["check-manifest"],
          "test": ["pytest", "pytest-cov"],
      },
      # entry_points={  # Optional
      #     "console_scripts": [
      #         "torus=scripts.torus:main",
      #     ],
      # },
      project_urls={  # Optional
          "Bug Reports": "https://github.com/JWKennington/torus/issues",
          "Funding": "https://www.buymeacoffee.com/locallytrivial",
          "Source": "https://github.com/JWKennington/torus",
      },
      )
